#######################################################################
#
# actor.py - The Actor class and its descendants, Player and Enemy.
#
# The Actor class provides movement and self-awareness code for the
# movable entities in the game (the player and enemies).  Derived
# classes Player and Enemy are used to customize this behaviour.
#
# PyRunner 0.2 (beta) - July 9th, 2003
# Copyright 2003 Ketche
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA
#
#######################################################################

# Python imports.
import random

# pygame imports.
import pygame.time

# PyRunner imports.
from const import *

# =====================================================================

# Class definitions.

# ---------------------------------------------------------------------
# Actor - A generic movable entity.
#
class Actor:
    # Constructor.
    # Parameters:  y - Row to start on.
    #              x - Column to start on.
    #
    def __init__(self, y=0, x=0):
        # Objects that can't be dug underneath.
        self.nodigunderlist = [FLOOR, PERMFLOOR, LADDER, POLE, GATELEFT,
                               GATERIGHT, SWITCH, SPIKES, EXITPOINT]
        # Objects that will hold up a player.
        self.supportivelist = [FLOOR, PERMFLOOR, LADDER, POLE]
        # Objects that a player can hang from if there is no support
        # underneath.
        self.hangfromlist = [LADDER, POLE, BAR]
        self.posy = y        # The actor's current row.
        self.posx = x        # The actor's current column.
        self.ofsx = 0        # Horiz. offset for incremental movement.
        self.ofsy = 0        # Vert. offset for incremental movement.
        self.moving = 0      # Flag if we are moving to a new square.
        self.lastaction = ACTIONSTILL  # Most recently taken action.
    # -----------------------------------------------------------------
    # canmoveto() - Tests if a move is allowed to the given space.
    # Parameters:  y - Row to move to.
    #              x - Column to move to.
    #              horizdir - Horizontal direction of the move.
    # Returns: 1 if the move is allowed, 0 if it is not.
    # Notes:  y and x will have wrapping applied.
    #
    def canmoveto(self, y, x, horizdir=DIRNONE):
        tmp = gamemap.wrapcoords(y, x)
        if tmp:
            y, x = tmp
            # Can't move to a spot where an enemy is.
            for enemy in enemylist:
                if (y, x) == (enemy.posy, enemy.posx):
                    return 0
            # Otherwise, can always move to a false square.
            if (y, x) in gamemap.false:
                return 1
        else:
            # Can't move off the screen.
            return 0
        sq = gamemap.getsquare(y, x)   # y, x are now wrapped
        # Can't walk through floors.
        if sq == FLOOR or sq == PERMFLOOR:
            return 0
        # Test for one-way gates.
        if horizdir == DIRLEFT:
            if sq == GATERIGHT:
                return 0
        if horizdir == DIRRIGHT:
            if sq == GATELEFT:
                return 0
        # All clear!
        return 1

    # -----------------------------------------------------------------
    # creditgold() - Tests if the actor has picked up a gold piece.
    # Notes:  Takes no action by default.  This method is intended to
    #         be overridden by derived classes.
    #
    def creditgold(self):
        pass

    # -----------------------------------------------------------------
    # canclimb() - Tests if the actor may climb from a given square.
    # Parameters:  y - Row to climb from.
    #              x - Column to climb from.
    # Notes:  y and x will have wrapping applied.
    #
    def canclimb(self, y, x):
        coords = gamemap.wrapcoords(y, x)
        if coords:
            # Can only climb if the square is a non-false ladder.
            if gamemap.data[coords[0]][coords[1]] == LADDER and\
               coords not in gamemap.false:
                return 1
        # Can't climb off screen
        return 0

    # -----------------------------------------------------------------
    # creditgold() - Tests if the actor has picked up a gold piece.
    # Notes:  Takes no action by default.  This method is intended to
    #         be overridden by derived classes.
    #
    def creditgold(self):
        pass

    # -----------------------------------------------------------------
    # canclimb() - Tests if the actor may climb from a given square.
    # Parameters:  y - Row to climb from.
    #              x - Column to climb from.
    # Notes:  y and x will have wrapping applied.
    #
    def canclimb(self, y, x):
        coords = gamemap.wrapcoords(y, x)
        if coords:
            # Can only climb if the square is a non-false ladder.
            if gamemap.data[coords[0]][coords[1]] == LADDER and\
               coords not in gamemap.false:
                return 1
        # Can't climb off screen
        return 0

    # -----------------------------------------------------------------
    # checkfall() - Tests if the actor should fall (using current pos.)
    # Notes:  Does not return a value.  Instead, self.falling will be
    #         set appropriately.
    #
    def checkfall(self):
        y, x = gamemap.wrapcoords(self.posy, self.posx)
        self.falling = 0
        fallcoords = gamemap.wrapcoords(y + 1, x)
        if not fallcoords:
            return
        for enemy in enemylist:
            if enemy.posy == fallcoords[0] and enemy.posx == fallcoords[1]:
                return
        sq = gamemap.getsquare(y, x)
        if sq not in self.hangfromlist or (y, x) in gamemap.false:
            if gamemap.getsquare(fallcoords[0], fallcoords[1]) not in\
               self.supportivelist or fallcoords in gamemap.false:
                self.falling = 1

    # -----------------------------------------------------------------
    # moveleft() - Moves the actor to the left.
    # Returns:  1 if the move was successful, 0 if not.
    #
    def moveleft(self):
        if self.canmoveto(self.posy, self.posx - 1, DIRLEFT):
            self.posx -= 1
            self.posy, self.posx = gamemap.wrapcoords(self.posy, self.posx)
            self.lastaction = ACTIONLEFT
            self.checkfall()
            return 1
        self.checkfall()
        return 0

    # -----------------------------------------------------------------
    # moveright() - Moves the actor to the right.
    # Returns:  1 if the move was successful, 0 if not.
    #
    def moveright(self):
        if self.canmoveto(self.posy, self.posx + 1, DIRRIGHT):
            self.posx += 1
            self.posy, self.posx = gamemap.wrapcoords(self.posy, self.posx)
            self.lastaction = ACTIONRIGHT ####### Don't set here!!!!
            self.checkfall()
            return 1
        self.checkfall()
        return 0

    # -----------------------------------------------------------------
    # moveup() - Moves the actor upwards.
    # Returns:  1 if the move was successful, 0 if not.
    #
    def moveup(self):
        if self.canclimb(self.posy, self.posx):
            if self.canmoveto(self.posy - 1, self.posx):
                self.posy -= 1
                self.posy, self.posx = gamemap.wrapcoords(self.posy, self.posx)
                self.lastaction = ACTIONUP ####### Don't set here!!!!
                self.checkfall()
                return 1
        self.checkfall()
        return 0

    # -----------------------------------------------------------------
    # movedown() - Moves the actor downwards.
    # Returns:  1 if the move was successful, 0 if not.
    #
    def movedown(self):
        if self.canmoveto(self.posy + 1, self.posx):
            self.posy += 1
            self.posy, self.posx = gamemap.wrapcoords(self.posy, self.posx)
            self.lastaction = ACTIONDOWN ####### Don't set here!!!!
            self.checkfall()
            return 1
        self.checkfall()
        return 0

    # -----------------------------------------------------------------
    # digleft() - Digs to the left of the actor.
    # Parameters:  framecount - The current frame counter.
    # Returns:  1 if the move was successful, 0 if not.
    # Notes:  Also used for switches.
    #
    def digleft(self, framecount):
        ## Much of this code is duplicated in digright(), which is a
        ## Bad Thing of course.  Centralize to a common function.

        if self.falling:
            self.lastaction = ACTIONDIGLEFT
            return

        # We need to know the spot the actor will dig underneath of
        # plus the spot actually being dug.
        leftcoords = gamemap.wrapcoords(self.posy, self.posx - 1)
        downleftcoords = gamemap.wrapcoords(self.posy + 1, self.posx - 1)
        if not leftcoords:
            # Can't dig off-screen.
            self.checkfall()
            return 0
        # Check if the dig action will flip a switch instead.
        if gamemap.data[leftcoords[0]][leftcoords[1]] == SWITCH and\
           leftcoords not in gamemap.false:
            gamemap.doswitch()
            self.checkfall()
            return 1
        # Only the player can actually dig a floor.
        if self is not player:
            # If we get here then somehow this method was called for
            # an enemy without a switch present.
            print "Bug: Enemy tried to dig (left)"
            return 0
        # Can't dig off the bottom of the screen or if we're past our
        # limit.
        if not downleftcoords or self.digsleft == 0:
            self.checkfall()
            return 0
        # Can only dig under spaces we can move to, unless it's because
        # there's an enemy there (enemies can be dug under).
        if self.canmoveto(self.posy, self.posx - 1) or\
           (self.posy, self.posx - 1) in\
           [(x.posy, x.posx) for x in enemylist]:
            # Ensure the square is diggable and the square above it does
            # not block digging (e.g. ladder).
            if (gamemap.getsquare(self.posy + 1, self.posx - 1) == FLOOR and\
               downleftcoords not in gamemap.false and\
               gamemap.getsquare(self.posy, self.posx - 1) not in\
               self.nodigunderlist) or\
               leftcoords in gamemap.false:
                # Dig it!
                gamemap.data[downleftcoords[0]][downleftcoords[1]] = EMPTY
                gamemap.diggedfloors.append((downleftcoords[0],
                                             downleftcoords[1],
                                             framecount))
                gamemap.draw(downleftcoords[0], downleftcoords[1])
                self.digsleft -= 1

    # -----------------------------------------------------------------
    # digright() - Digs to the right of the actor.
    # Parameters:  framecount - The current frame counter.
    # Returns:  1 if the move was successful, 0 if not.
    # Notes:  Also used for switches.
    #
    def digright(self, framecount):
        # This method is virtually identical to digleft().
        if self.falling:
            self.lastaction = ACTIONDIGRIGHT
            return
        rightcoords = gamemap.wrapcoords(self.posy, self.posx + 1)
        downrightcoords = gamemap.wrapcoords(self.posy + 1, self.posx + 1)
        if not rightcoords:
            self.checkfall()
            return 0
        if gamemap.data[rightcoords[0]][rightcoords[1]] == SWITCH and\
           rightcoords not in gamemap.false:
            gamemap.doswitch()
            self.checkfall()
            return 1
        if self is not player:
            print "Bug: Enemy tried to dig (right)"
            return 0
        if not downrightcoords or self.digsleft == 0:
            self.checkfall()
            return 0
        if self.canmoveto(self.posy, self.posx + 1) or\
           (self.posy, self.posx + 1) in\
           [(x.posy, x.posx) for x in enemylist]:
            if (gamemap.getsquare(self.posy + 1, self.posx + 1) == FLOOR and\
               downrightcoords not in gamemap.false and\
               gamemap.getsquare(self.posy, self.posx + 1) not in\
               self.nodigunderlist) or\
               rightcoords in gamemap.false:
                gamemap.data[downrightcoords[0]][downrightcoords[1]] = EMPTY
                gamemap.diggedfloors.append((downrightcoords[0],
                                             downrightcoords[1],
                                             framecount))
                gamemap.draw(downrightcoords[0], downrightcoords[1])
                self.digsleft -= 1

    # -----------------------------------------------------------------
    # domove() - Chooses the next move for the actor.
    # Notes:  Takes no action by default.  This method is intended to
    #         be overridden by derived classes.
    #
    def domove(self):
        pass

    # -----------------------------------------------------------------
    # checkdeath() - Test if this actor is about to die.
    # Returns:  1 if death is imminent, 0 otherwise.
    #
    def checkdeath(self):
        # Can't die in a false square.
        if (self.posy, self.posx) in gamemap.false:
            return 0
        # We die if we hit spikes or if a floor closes over us.
        sq = gamemap.data[self.posy][self.posx]
        if sq == FLOOR or sq == SPIKES:
            self.lastaction = ACTIONDEATH
            return 1
        return 0

    # -----------------------------------------------------------------
    # checkexit() - Checks if the actor has exited the level.
    # Returns: A two-tuple containing the coordinates where the exit
    #          took place, None otherwise.
    #
    ## This is in the general Actor class because in the future it may
    ## be neat to allow enemies to exit to the next level with the
    ## player.
    ## The return value is not used yet; it is intended for a future
    ## enhancement that would send the player to different levels based
    ## on where they exited.
    def checkexit(self):
        # Can't exit until all stages completed.
        if gamemap.stage != gamemap.numstages:
            return
        # Test the exit points.
        for exitsq in gamemap.exits:
            if exitsq[0] == self.posy and exitsq[1] == self.posx:
                return (self.posy, self.posx)
        # Test if at the top row (assuming that option is enabled).
        if gamemap.exitattop and self.posy == 0:
            # We treat the top as a single large "exit point" so
            # the -1, -1 location is a special code signifying this.
            return (-1, -1)
        return None

# ---------------------------------------------------------------------
# Player - The actor under human control.
#
class Player(Actor):
    # Constructor.
    # Parameters:  y - Row to start on.
    #              x - Column to start on.
    #
    def __init__(self, y=0, x=0):
        Actor.__init__(self, y, x)     # Initialize the superclass.
        self.score = 0                 # The game score.
        self.lives = STARTLIVES        # How cat-like the player is.
        self.digsleft = gamemap.diglimit    # No. of simultaneous digs.

    # -----------------------------------------------------------------
    # creditgold() - Tests if the actor has picked up a gold piece.
    # Overrides Method In:  Actor
    #
    def creditgold(self):
        if gamemap.data[self.posy][self.posx] == GOLD and\
           (self.posy, self.posx) not in gamemap.false:
            # Credit the gold to the player's score.
            gamemap.data[self.posy][self.posx] = EMPTY
            if gamemap.stage == gamemap.numstages:
                self.score += BONUSGOLDSCORE
            else:
                self.score += GOLDSCORE
            gamemap.goldleft -= 1
            # Erase the gold piece.  We have to do this here so it's
            # clear that the player has picked it up.
            gamemap.draw(self.posy, self.posx)
            if gamemap.goldleft == 0:
                gamemap.doafter(enemylist)

    # -----------------------------------------------------------------
    # checkdeath() - Test if this actor is about to die.
    # Returns:  1 if death is imminent, 0 otherwise.
    # Overrides Method In:  Actor
    #
    def checkdeath(self):
        if Actor.checkdeath(self):
            return 1
        # The player can also die if touched by an enemy.
        for enemy in enemylist:
            if (enemy.posy, enemy.posx) == (self.posy, self.posx):
                return 1
        return 0

# ---------------------------------------------------------------------
# Enemy - A computer-controlled actor.
#
class Enemy(Actor):
    # Constructor.
    # Parameters:  y - Row to start on.
    #              x - Column to start on.
    # Overrides Method In:  Actor
    #
    def __init__(self, y=0, x=0):
        Actor.__init__(self, y, x)     # Initialize the superclass.
        self.hasgold = 0     # Flag if the enemy is carrying gold.
        self.lastmovedir = DIRNONE     # The direction of our last move.
        self.justpulledswitch = 0      # Flag if we just hit a switch.
        self.justdroppedgold = 0       # Flag if we just dropped a gold.

    # -----------------------------------------------------------------
    # creditgold() - Tests if the actor has picked up a gold piece.
    # Overrides Method In:  Actor
    #
    def creditgold(self):
        # Enemies carry around the gold so the player can't get it.
        # This is always called after the enemy's move so we must
        # be sure if we just dropped gold that we aren't going to
        # pick it up again.
        if gamemap.data[self.posy][self.posx] == GOLD and\
           (self.posy, self.posx) not in gamemap.false and\
           not self.hasgold and not self.justdroppedgold:
            gamemap.data[self.posy][self.posx] = EMPTY
            self.hasgold = 1

    # -----------------------------------------------------------------
    # checkfall() - Tests if the actor should fall (using current pos.)
    # Notes:  Does not return a value.  Instead, self.falling will be
    #         set appropriately.
    # Overrides Method In:  Actor
    #
    def checkfall(self):
        Actor.checkfall(self)
        # Enemies always drop their gold when they are falling.  This
        # is a traditional behaviour.
        if self.hasgold and self.falling:
            self.dropgold()

    # -----------------------------------------------------------------
    # dropgold() - Tells the actor to drop any gold being held.
    #
    def dropgold(self):
        # We can only drop gold if the space is empty.
        if self.hasgold and\
           gamemap.data[self.posy][self.posx] == EMPTY:
            gamemap.data[self.posy][self.posx] = GOLD
            self.hasgold = 0
            self.justdroppedgold = 1

    # -----------------------------------------------------------------
    # checkdeath() - Test if this actor is about to die.
    # Returns:  1 if death is imminent, 0 otherwise.
    # Overrides Method In:  Actor
    #
    def checkdeath(self):
        dead = Actor.checkdeath(self)
        # If all goes well, enemies can't die with their gold, but
        # there are cases where they can't get rid of it so we make
        # sure to remove it from the game if it's not dropped.
        if dead and self.hasgold:
            self.lastaction = ACTIONDEATH
            gamemap.goldleft -= 1
            if gamemap.goldleft == 0:
                gamemap.doafter(enemylist)
        return dead

    # -----------------------------------------------------------------
    # domove() - Chooses the next move for the actor.
    # Overrides Method In:  Actor
    #
    def domove(self):
        gamemap.draw(self.posy, self.posx)
        if self.falling:
            # Can't move if we're falling.
            self.movedown()
        else:
            ## This code lets the enemy home in on the player.
            ## As it is it makes the enemies WAY TOO SMART.  (I have
            ## noticed that this is often much easier to do than
            ## making enemies only semi-smart.)
            ## Could be used in an "expert" mode sometime.
            #didmove = 0
            #if not didmove and player.posx < self.posx:
            #    didmove = self.moveleft()
            #if not didmove and player.posx > self.posx:
            #    didmove = self.moveright()
            #if not didmove and player.posy < self.posy:
            #    didmove = self.moveup()
            #if not didmove and player.posy > self.posy:
            #    didmove = self.movedown()
            #if not didmove:
            #    dir = random.randrange(DIRSTART, DIREND)
            #    if dir == DIRLEFT: self.moveleft()
            #    if dir == DIRRIGHT: self.moveright()
            #    if dir == DIRUP: self.moveup()
            #    if dir == DIRDOWN: self.movedown()

            res = 0          # Flag that holds if a move was successful.
            # Try to pull a switch, but only if we just didn't already
            # do so.  (If we didn't have this check, an enemy could
            # keep pulling the same switch over and over again.)
            if not self.justpulledswitch:
                tmpcoords = gamemap.wrapcoords(self.posy, self.posx - 1)
                if tmpcoords:
                    if gamemap.data[tmpcoords[0]][tmpcoords[1]] == SWITCH and\
                       tmpcoords not in gamemap.false and\
                       random.randrange(MAXCHANCE) < ENEMYSWITCHCHANCE:
                        res = self.digleft(0)
                # In case there happen to be switches on both sides of
                # the enemy, we make sure to pull only one.
                if not res:
                    tmpcoords = gamemap.wrapcoords(self.posy, self.posx + 1)
                    if tmpcoords:
                        if gamemap.data[tmpcoords[0]][tmpcoords[1]] ==\
                           SWITCH and\
                           tmpcoords not in gamemap.false and\
                           random.randrange(MAXCHANCE) < ENEMYSWITCHCHANCE:
                            res = self.digright(0)

            # For enemies, pulling the switch is considered a whole
            # move.  (The player can pull/dig and move simultaneously.)
            if not res:
                # Main move logic.
                ## The enemies are way too brain-dead.  They need to
                ## be aware of their immediate surroundings and also
                ## where the player is.
                ## Also, to be totally true to tradition, there should
                ## be no situations in which enemies stand still.
                sq = gamemap.data[self.posy][self.posx]

                # This bit chooses a new direction.  If the player just
                # moved onto a ladder or pole we also force them to
                # consider a direction change so they can climb/descend
                # it.
                if self.lastmovedir == DIRNONE or\
                   ((self.lastmovedir == DIRLEFT or\
                   self.lastmovedir == DIRRIGHT) and\
                   (sq == LADDER or sq == POLE)):
                    self.lastmovedir = random.randrange(DIRSTART, DIREND)

                # This is the only bit of player awareness the enemies
                # have.  If a player is on a ladder/pole and they
                # are on the same row as the player, there's a chance
                # they will change direction, though WHICH direction
                # is random.
                # NOTE:  This code is actually needed so enemies don't
                #        stay permanently on the ladders on the main
                #        adventure level "Nuts" and others like it.
                if random.randrange(MAXCHANCE) < ENEMYJUMPOFFCHANCE:
                    if (self.lastmovedir == DIRUP or\
                       self.lastmovedir == DIRDOWN) and\
                       player.posy == self.posy:
                        self.lastmovedir = random.randrange(DIRSTART, DIREND)

                # The basic principle behind enemy movement is that
                # enemies always continue in the direction they were
                # just moving until they can't anymore.
                if self.lastmovedir == DIRLEFT: res = self.moveleft()
                if self.lastmovedir == DIRRIGHT: res = self.moveright()
                if self.lastmovedir == DIRUP: res = self.moveup()
                if self.lastmovedir == DIRDOWN: res = self.movedown()
                if not res:
                    self.lastmovedir = DIRNONE
                else:
                    self.justpulledswitch = 0
                    self.justdroppedgold = 0
            else:
                self.justpulledswitch = 1
                self.justdroppedgold = 0

        self.creditgold()
        # Enemies traditionally have a small chance of randomly
        # dropping their gold.
        if self.hasgold and\
           random.randrange(MAXCHANCE) < ENEMYRANDDROPCHANCE:
            self.dropgold()

# End of file:  actor.py
