#######################################################################
#
# map.py - The Map class.
#
# The Map class holds the inanimate objects in the game.
#
# PyRunner 0.2 (beta) - July 9th, 2003
# Copyright 2003 Ketche 
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA
#
#######################################################################

# Python imports.
import os.path

# pygame imports.
import pygame.image
from pygame.locals import *

# PyRunner imports.
import actor
from const import *

# =====================================================================

# Class definitions.

# ---------------------------------------------------------------------
# Map - Keeps track of on-screen objects.
# Notes:  All "coordinates" in lists are two-tuples:  (y, x)
#
class Map:
    # Symbols for normal visible objects.
    symbols = {'.': EMPTY, '#': FLOOR, '%': PERMFLOOR, '=': LADDER,
               '-': BAR, '|': POLE, '$': GOLD, '<': GATELEFT,
               '>': GATERIGHT, '/': SWITCH, '^': SPIKES, '0': EXITPOINT}
    # Symbols for invisible objects.
    specialsymbols = {'@': PLAYERSTART, '!': ENEMYSTART,
                      '?': ENEMYREGEN}
    # Symbols for objects that are shown but don't exist.
    falsesymbols = {'+': FLOOR, '*': PERMFLOOR, '"': LADDER,
                    "'": BAR, ':': POLE, '&': SPIKES, '_': EXITPOINT,
                    '~': GOLD, '\\': SWITCH}
    # Symbols for objects which appear after stage 1.
    aftersymbols1 = {'F': FLOOR, 'M': PERMFLOOR, 'L': LADDER,
               'B': BAR, 'P': POLE, 'G': GOLD, '[': GATELEFT,
               ']': GATERIGHT, 'W': SWITCH, 'S': SPIKES,
               'E': ENEMYSTART, 'R': ENEMYREGEN, 'X': EXITPOINT}
    # Symbols for objects which appear after stage 2.
    aftersymbols2 = {'f': FLOOR, 'm': PERMFLOOR, 'l': LADDER,
               'b': BAR, 'p': POLE, 'g': GOLD, '(': GATELEFT,
               ')': GATERIGHT, 'w': SWITCH, 's': SPIKES,
               'e': ENEMYSTART, 'r': ENEMYREGEN, 'x': EXITPOINT}
    # Tuple containing all after-stage symbols.
    aftersymbols = (aftersymbols1, aftersymbols2)

    # Constructor.
    # Parameters:  screen - The screen display surface.
    #              mapfile - Name of the file to load the map from.
    #                        If omitted, a very basic map is created.
    #
    def __init__(self, screen, mapfile=None):
        # The list of sprite files.
        ## Not sufficient for animation, e.g. visual switches.
        self.spritedatafiles = ("empty.bmp", "floor.bmp", "permfloor.bmp",
            "ladder.bmp", "bar.bmp", "pole.bmp", "gold.bmp",
            "gateleft.bmp", "gateright.bmp", "switch.bmp", "spikes.bmp",
            "exitpoint.bmp", "logo.bmp")
        self.data = []                 # The actual map object data.
        self.playerstart = ()          # Player's starting coordinates.
        self.enemystart = []           # Enemy start coords.
        self.enemyregen = []           # Enemy regen. point coords.
        self.exits = []                # Exit point coords.
        self.false = []                # False object coords.
        # Each sublist of after is filled with a tuple:  (y, x, object)
        self.after = [[], []]          # After-stage objects.
        self.hints = []                # Hint strings.
        self.meta = ""                 # Authorship information.
        self.numstages = 1             # No. of gold-collecting stages.
        self.diglimit = 99             # Maximum simultaneous digs.
        self.exitattop = 1             # Flag if can exit on row 0.
        self.flipgatesatend = 0        # Flag:  Flip gates after stages.
        self.horizwrap = 0             # Flag:  Wrap around sides.
        self.vertwrap = 0              # Flag:  Wrap top/bottom.
        self.artwork = "default"       # Name of artwork set to use.
        self.goldleft = 0              # Gold remaining in this stage.
        self.gates = []                # One-way gate coordinates.
        self.oddflips = 0         # Flag:  Gates not in start state.
        self.stage = 0                 # Current stage.
        # The diggedfloors list contains tuples of the form:
        # (y, x, frame) where frame is the frame number when the dig
        # took place.
        self.diggedfloors = []         # Floors currently dug out.
        ## Could (and probably should) get this from pygame.
        self.screen = screen           # The display surface.
        # This is the mutable option list.  Once this is finalized
        # it is converted to a string and stored in optstr.
        ## This is awkward and kind of lame - find a better way.
        self.optlist = [' ', ' ', 'T', ' ']
        # The finalized option string.
        self.optstr = "  T "
        tmp = []             # Used to fill up the map.
        # Fill up the entire map with empty spaces.
        ## Currently does nothing since the default sizes are 0.
        for i in range(MAPWIDTH):
            tmp.append(EMPTY)
        for i in range(MAPHEIGHT):
            # Here we make a copy of tmp for each row.  Otherwise,
            # changing one row would change them all since they'd
            # all refer to the same object.
            self.data.append(list(tmp))
        if mapfile:          # Load from map data file.
            self.initfromfile(mapfile)
        self.sprites = []    # Object sprite surfaces.
        # Load the object sprites.
        for filename in self.spritedatafiles:
            surf = pygame.image.load(os.path.join(self.artwork, filename))
            #Err
            surf.set_colorkey(TRANSTUPLE, RLEACCEL)
            self.sprites.append(surf.convert())

    # -----------------------------------------------------------------
    # initfromfile() - Loads map data from a file.
    # Parameters:  filename - The name of the map data file.
    #
    def initfromfile(self, filename):
        # Helper function that determines if an option is to be set on
        # or off.
        #
        def istrue(str):
            # Any of these strings will be considered to mean "on",
            # and everything else means "off".
            truenames = ("true", "1", "on", "yes")
            return str.lower() in truenames

        maprow = 0        # Next "map" line will define this row.
        f = open(filename, "r")
        #Err
        while 1:
            str = f.readline()
            if str == "":           # Test for end of file.
                break
            if str[0] == ';':       # Ignore comments.
                continue
            if str[-1] == '\n':     # Strip newline, if any.
                str = str[:-1]
            if str.strip() == "":   # Ignore blank lines.
                continue

            # We want just one split so filename specifications with
            # spaces (under Windows) won't be butchered.
            tmp = str.split(None, 1)
            # Valid lines always have two
            if len(tmp) != 2:
                #Err - report a bad line was found
                continue
            # The first word is the keywords, and the rest of the string
            # is considered to be the options for it.
            keyword = tmp[0].lower()
            options = tmp[1].strip()
            if keyword == "map":
                if maprow == MAPHEIGHT:
                    #Err - report too many map lines found
                    pass
                for mapcol in range(MAPWIDTH):
                    if mapcol == len(options):
                        #Err - report line too long
                        break
                    # Read in the map symbol by seeing if it's in one
                    # of the dictionaries.
                    if options[mapcol] in self.symbols.keys():
                        self.data[maprow][mapcol] =\
                            self.symbols[options[mapcol]]
                        if self.data[maprow][mapcol] == GOLD:
                            self.goldleft += 1
                        if self.data[maprow][mapcol] == GATELEFT or\
                           self.data[maprow][mapcol] == GATERIGHT:
                            self.gates.append((maprow, mapcol))
                        if self.data[maprow][mapcol] == EXITPOINT:
                            self.exits.append((maprow, mapcol))
                    elif options[mapcol] in self.specialsymbols.keys():
                        meaning = self.specialsymbols[options[mapcol]]
                        if meaning == PLAYERSTART:
                            if self.playerstart != ():
                                #Err - only one playerstart allowed
                                pass
                            self.playerstart = (maprow, mapcol)
                        elif meaning == ENEMYSTART:
                            self.enemystart.append((maprow, mapcol))
                        elif meaning == ENEMYREGEN:
                            self.enemyregen.append((maprow, mapcol))
                        else:
                            #Err - somehow the dict got screwed up
                            pass
                        self.data[maprow][mapcol] = EMPTY
                    elif options[mapcol] in self.falsesymbols.keys():
                        self.false.append((maprow, mapcol))
                        self.data[maprow][mapcol] =\
                            self.falsesymbols[options[mapcol]]
                    elif options[mapcol] in self.aftersymbols[0].keys():
                        self.after[0].append((maprow, mapcol,
                            self.aftersymbols[0][options[mapcol]]))
                        self.data[maprow][mapcol] = EMPTY
                    elif options[mapcol] in self.aftersymbols[1].keys():
                        self.after[1].append((maprow, mapcol,
                            self.aftersymbols[1][options[mapcol]]))
                        self.data[maprow][mapcol] = EMPTY
                    else:
                        #Err - unknown symbol
                        pass
                maprow += 1
            elif keyword == "meta":
                self.meta = options
            elif keyword == "numstages":
                self.numstages = int(options)
            elif keyword == "diglimit":
                self.diglimit = int(options)
            elif keyword == "exitattop":
                self.exitattop = istrue(options)
                if self.exitattop:
                    ## optlist indexes should have named constants.
                    self.optlist[2] = 'T'
                else:
                    self.optlist[2] = ' '
            elif keyword == "flipgates":
                self.flipgatesatend = istrue(options)
                if self.flipgatesatend:
                    self.optlist[3] = 'F'
                else:
                    self.optlist[3] = ' '
            elif keyword == "horizwrap":
                self.horizwrap = istrue(options)
                if self.horizwrap:
                    self.optlist[0] = 'H'
                else:
                    self.optlist[0] = ' '
            elif keyword == "vertwrap":
                self.vertwrap = istrue(options)
                if self.vertwrap:
                    self.optlist[1] = 'V'
                else:
                    self.optlist[1] = ' '
            elif keyword == "artwork":
                self.artwork = os.path.normpath(options)
            elif keyword == "hint":
                self.hints.append(options)
            else:
                #Err - invalid keyword
                pass
        f.close()

        # Convert the option list to a string.
        s = ""
        for opt in self.optlist:
            s += opt
        self.optstr = s

    # -----------------------------------------------------------------
    # wrapcoords() - Ensures coords are in the map, wrapping if needed.
    # Returns:  Tuple containing the adjusted (y, x), or None if the
    #           coordinates cannot be wrapped because wrapping is
    #           disabled.
    #
    def wrapcoords(self, y, x):
        # This is a simple function.  If coordinates are out of range,
        # wrap them in the appropriate direction so they are (if
        # wrapping is enabled) or (if this can't be done) abort.
        ## Doesn't handle coordinates that are more than one range
        ## out-of-range.  Not a problem in the current implementation
        ## as coords. can only be one space out at most anyhow, but
        ## future changes could break this.
        if y < 0:
            if not self.vertwrap:
                return None
            y += MAPHEIGHT
        if y >= MAPHEIGHT:
            if not self.vertwrap:
                return None
            y -= MAPHEIGHT
        if x < 0:
            if not self.horizwrap:
                return None
            x += MAPWIDTH
        if x >= MAPWIDTH:
            if not self.horizwrap:
                return None
            x -= MAPWIDTH
        return (y, x)

    # -----------------------------------------------------------------
    # getsquare() - Gets a square of map data, wrapping if needed.
    # Parameters:  y - Row of the square to get.
    #              x - Column of the square to get.
    # Notes:  Only use this when wrapping is absolutely needed and when
    #         it hasn't been done in advance for other reasons.  Other-
    #         wise, it's faster to access the data member directly.
    def getsquare(self, y, x):
        tmp = self.wrapcoords(y, x)
        if not tmp:
            return PERMFLOOR
        return self.data[tmp[0]][tmp[1]]

    # -----------------------------------------------------------------
    # doafter() - Perform required actions when a stage is completed.
    # Parameters:  enemylist - The game's list of Enemy objects.
    #
    def doafter(self, enemylist):
        # Make sure there's still a stage to go.
        if self.stage != self.numstages:
            # Flip gates after the last stage, if needed.
            if self.flipgatesatend and\
               self.stage == self.numstages - 1:
                self.doswitch()

            # Add all of the after-stage objects.
            for item in self.after[self.stage]:
                self.data[item[0]][item[1]] = item[2]
                if item[2] == GOLD:
                    self.goldleft += 1
                if item[2] == GATELEFT or item[2] == GATERIGHT:
                    self.gates.append((item[0], item[1]))
                if item[2] == EXITPOINT:
                    self.exits.append((item[0], item[1]))
                if item[2] == ENEMYSTART:
                    self.enemystart.append((item[0], item[1]))
                    enemylist.append(actor.Enemy(item[0], item[1]))
                    self.data[item[0]][item[1]] = EMPTY
                if item[2] == ENEMYREGEN:
                    self.enemyregen.append((item[0], item[1]))
                    self.data[item[0]][item[1]] = EMPTY
            self.stage += 1
            self.draw()

    # -----------------------------------------------------------------
    # doswitch() - Flips a switch to change the one-way gates.
    #
    def doswitch(self):
        for gate in self.gates:
            assert self.data[gate[0]][gate[1]] in (GATELEFT, GATERIGHT)
            if self.data[gate[0]][gate[1]] == GATELEFT:
                self.data[gate[0]][gate[1]] = GATERIGHT
            else:
                self.data[gate[0]][gate[1]] = GATELEFT
            self.draw(gate[0], gate[1])
        self.oddflips = not self.oddflips

    # draw() - Draws a portion of the map on the screen.
    # Parameters:  y1 - Starting row to draw (inclusive).
    #              x1 - Stating column to draw (inclusive).
    #              y2 - Ending row (exclusive).
    #              x2 - Ending column (exclusive).
    # Notes:  May be called with 0, 2, or 4 parameters.  0 will redraw
    #         the entire map.  2 will redraw only the given square.
    #         4 redraws the specified region.
    def draw(self, y1 = None, x1 = None, y2 = None, x2 = None):
        if not y1 or not x1:            # Draw whole map.
            y1 = 0
            x1 = 0
            y2 = MAPHEIGHT
            x2 = MAPWIDTH
        elif not y2 or not x2:          # Draw one square.
            y2 = y1 + 1
            x2 = x1 + 1
        else:                           # Draw specified region.
            # Check to make sure we're inbounds.
            if y1 < 0: y1 = 0
            if x1 < 0: x1 = 0
            if y2 > MAPHEIGHT: y2 = MAPHEIGHT
            if x2 > MAPWIDTH: x2 = MAPWIDTH

        # "Negative" (reversed) regions aren't allowed.
        if y1 > y2 or x1 > x2: return

        self.screen.fill((0, 0, 0),
                         (x1 * SPRITEWIDTH, y1 * SPRITEHEIGHT,
                          (x2 - x1) * SPRITEWIDTH, (y2 - y1) * SPRITEHEIGHT))
        y = y1 * SPRITEHEIGHT
        for i in range(y1, y2):
            x = x1 * SPRITEWIDTH
            for j in range(x1, x2):
                self.screen.blit(self.sprites[self.data[i][j]], (x, y))
                x += SPRITEWIDTH
            y += SPRITEHEIGHT

# End of file:  map.py
