#######################################################################
#
# pyrunner.py - PyRunner startup/shutdown code.  (Runnable)
#
# The main PyRunner program (i.e. the stuff that knits the other parts
# together.  This is the file that should be executed in order to play
# PyRunner.
#
# PyRunner 0.2 (beta) - July 9th, 2003
# Copyright 2003 Ketche
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330,
# Boston, MA  02111-1307  USA
#
#######################################################################
#
# PyRunner Source Notes
# ---------------------
#
# Comments beginning with # are normal (documentary) comments.
# Comments beginning with ## are notes that indicate where things
# need to be improved.
# Comments of "#Err" indicate where proper error-handling (usually
# try/except blocks) needs to be implemented.
#
# All map positions used by PyRunner are reported in row, col order
# (y, x).  This is to make them consistent with the way they are stored
# in the Map class.  (Graphics positions like that used by pygame and
# BmpFont use x, y order of course.)
#
# Check routines like checkfall()/creditgold() are called way more
# often than they need to be.  This is laziness on my part and may
# be responsible for a bit of odd behaviour.
#
# I'm fairly new to Python so if something looks clumsy or awkward it's
# because I haven't learned the elegant way to do it in Python yet.
#
# I'm also a long-time C programmer whose brain still hasn't entirely
# shifted over to OOP yet.  There are lots of places where this makes
# things quite ugly, and where a pure OO solution would be cleaner.
# However, making all class variables "public" is an intentional
# design so the overhead of accessor functions can be avoided.  I find
# that with game programming, it helps to throw some of the OO rules
# out the window.
#
# The module bmpfont.py is a freeware module that was contributed to
# the pygame Code Repository (http://www.pygame.org/pcr/) by the
# author, so its use is not bound by the GPL.
#
#######################################################################

# Python imports.
import cPickle
import os.path
import random
import sys
import time

# pygame imports.
import pygame
import pygame.image
from pygame.locals import *

# Utility imports.
from bmpfont import BmpFont

# PyRunner imports.
from anim import *
import actor
from const import *
import highs
import map

# =====================================================================

# EXECUTION BEGINS HERE

# Setup the pygame engine.
pygame.init()
pygame.display.set_caption("PyRunner " + VERSION)
screen = pygame.display.set_mode((SCREENWIDTH, SCREENHEIGHT), FULLSCREEN, 16)
#Err
pygame.mouse.set_visible(0)
os.chdir("/usr/lib/pyrunner")

# PyRunner uses bitmapped fonts for text display.  Bitmapped fonts are
# faster (generally), easier to use because all characters are the same
# size, and more portable.  However, we're stuck with the one colour,
# but that's okay for PyRunner.
font = BmpFont("font.idx")             # Holds the bitmapped font.
#Err

# Title / name entry screen.
## Should be made into a function to allow multiple games in one
## session.
screen.fill((0, 0, 0))
font.blit("Enter your name:  ", screen, (0, 0))
npos = 0                     # Position within the name string.
# Strings aren't mutable so a list is used for the actual editing.
namechars = []               # The list of characters for the name.
playername = ""              # The name string.
pygame.key.set_repeat(250, 20)

# Key entry loop to get a string and display it with the bitmapped font.
## This should also be a function so it can be reused.
while 1:
    event = pygame.event.poll()
    if event.type == QUIT:
        sys.exit(0)
    if event.type == KEYDOWN:
        if event.key == K_ESCAPE:
            sys.exit(0)
        elif event.key == K_BACKSPACE and npos > 0:
            # Destructive backspace.
            npos -= 1
            screen.fill((0, 0, 0), ((18 + npos) * font.width, 0,
                                    font.width, font.height))
            namechars.pop()
        elif event.key == K_RETURN:
            # Done entry.
            if npos == 0:
                playername = "NokiRunner"
            else:
                playername = ""
                # Convert the list to a string.
                for ch in namechars:
                    playername += ch
            font.blit("Welcome, " + playername + "!", screen, (0, 2))
            pygame.display.flip()
            break
        else:
            ## Ugly.  pygame doesn't have a facility for plain-text
            ## input so we are stuck with using unshifted characters
            ## only.  Maybe stdin could be used for input here?
            # Look for standard, displayable ASCII characters.
            if event.key >= ord(' ') and event.key <= ord('~') and\
               npos < 30:
                namechars.append(chr(event.key).upper())
                font.blit(chr(event.key).upper(), screen, (18 + npos, 0))
                npos += 1
    pygame.display.flip()

# Main initialization code.

# We'll be using keystates, so auto-repeat is disabled.  This isn't
# necessary but probably is wise because players will be holding down
# keys all the time and this would generate a lot of events if repeat
# is enabled.
pygame.key.set_repeat()
playerdead = 0               # Whether the player has any lives left.
curlevel = 0                 # The current level number being played.
nextbonuslife = BONUSLIFESCORE        # Score needed for bonus life.
levelfiles = []              # Names of level files (from .gam file).
gamefilename = "pyrunner.gam"         # Name of adventure file.
line2str = ""                # The string on line 2 of the status bar.
skiplvl = 0

# Load a different adventure if one is specified on the command line.
if len(sys.argv) > 1:
    gamefilename = os.path.normpath(sys.argv[1])

# Read the adventure file, which is just a list of level file names.
f = open(gamefilename, "r")
#Err
for x in f.xreadlines():
    if x[-1] == '\n': x = x[:-1]
    levelfiles.append(os.path.normpath(x.strip()))
f.close()

# =====================================================================

# Function definitions

# ---------------------------------------------------------------------
# levelstartscreen() - The "get ready" screen for when a level starts.
#
def levelstartscreen():
    global player
    global skiplvl

    # This delay lets the player see where & how they died/exited.
    pygame.time.delay(1000)

    ## The screen coordinates here should be named constants so they're
    ## easier to change.
    screen.fill((0, 0, 0))
    tmpstr = "PyRunner " + VERSION + " - by Ketche, Original PyRunner by Paul Sidorsky"
    font.blit(tmpstr.center(80), screen, (0, 10))
    tmpstr = "Score: %07i      Lives: %i" % (player.score, player.lives)
    font.blit(tmpstr.center(80), screen, (0, 12))
    tmpstr = "Entering Level %i" % (curlevel)
    font.blit(tmpstr.center(80), screen, (0, 14))
    font.blit("Get Ready!".center(80), screen, (0, 16))
    font.blit("Play begins in 3 seconds", screen, (28, 18))

    skiplvl = 0

    # The countdown loop.
    for x in range(3, 0, -1):
        pygame.event.pump()
        event = pygame.event.poll()
        if event.type == KEYDOWN:
        	if event.key == K_UP:
			skiplvl = 1
        screen.fill((0, 0, 0), (43 * font.width, 18 * font.height,
                                font.width, font.height))
        font.blit(str(x), screen, (43, 18))
        pygame.display.flip()
        pygame.time.delay(1000)

#    if (skiplvl==1):
#        break

# ---------------------------------------------------------------------
# updateframe() - Updates the screen for each frame.
#
def updateframe():
    global nextbonuslife

    # Credit an extra life if the player's score is high enough.  This
    # is done here because there are several places where score can
    # be altered.
    ## It's a bad habit to put unnecessary checks in the main loop, as
    ## each slows down the game very slightly (they can pile up!).
    ## This check should ideally only be done when player.score is
    ## modified - likely via accessor methods, with nextbonuslife as a
    ## member of the Player class.
    if player.score >= nextbonuslife:
        player.lives += 1
        nextbonuslife += BONUSLIFESCORE

    # Draw the player and enemies.  Enemies get the higher z-order
    # position (i.e. they will overwrite the player) so it's clearer
    # when the player comes into contact with one.
    screen.blit(playeranims[player.lastaction][animcount],
                ((player.posx * SPRITEWIDTH) + player.ofsx,
                 (player.posy * SPRITEHEIGHT) + player.ofsy))
    for enemy in enemylist:
        screen.blit(enemyanims[enemy.lastaction][animcount],
                    ((enemy.posx * SPRITEWIDTH) + enemy.ofsx,
                     (enemy.posy * SPRITEHEIGHT) + enemy.ofsy))
    screen.fill((0, 0, 0),
                (0, VIEWHEIGHT,
                 VIEWWIDTH, SCREENHEIGHT - VIEWHEIGHT - 20))

    # Draw the status bar.
    ## Should only be updating the digits and not redrawing the whole
    ## bar, because each character is an expensive blit operation.
    font.blit("Score: %07i   Lives: %-2i   Digs: %-2i   Gold: %-3i   "
              "Stages: %-1i   Level: %-3i  %s" %\
              (player.score, player.lives, player.digsleft,
               gamemap.goldleft, gamemap.numstages - gamemap.stage,
               curlevel, gamemap.optstr),
              screen, (0, VIEWHEIGHT + 2), 0)
    pygame.display.flip()

# ---------------------------------------------------------------------
# checktimedevents() - The daemon that manages timed events.
#
def checktimedevents():
    ## To be 100% safe we should really adjust for get_ticks()
    ## wraparound if it occurs, even though it's highly unlikely.

    # Check to see if any digged floors are ready to regenerate.
    for floor in gamemap.diggedfloors:
        if framecount >= floor[2] + FLOORREGENDELAY:
            # If there happens to be a gold where the floor is supposed
            # to go, we need to remove it from the game or the stage
            # will be impossible to finish.
            if gamemap.data[floor[0]][floor[1]] == GOLD:
                gamemap.goldleft -= 1
                if gamemap.goldleft == 0:
                    gamemap.doafter(enemylist)
            gamemap.data[floor[0]][floor[1]] = FLOOR
            gamemap.draw(floor[0], floor[1])
            gamemap.diggedfloors.remove(floor)
            player.digsleft += 1

    # Check if dead enemies are ready to regenerate.
    for t in enemydeathtimes:
        if framecount >= t + ENEMYREGENDELAY:
            enemydeathtimes.remove(t)
            newpos = random.choice(gamemap.enemyregen)
            enemylist.append(actor.Enemy(newpos[0], newpos[1]))

# ---------------------------------------------------------------------
# printonline2() - Used to update line 2 of the status bar.
# Parameters:  msg - If specified, the string to display.  If omitted,
#                    the previous string will be displayed.  This is
#                    used to restore the current hint after a death.
#
def printonline2(msg=None):
    global line2str
    if not msg:
        msg = line2str
    screen.fill((0, 0, 0),
                (0, SCREENHEIGHT - 20, VIEWWIDTH, 20))
    font.blit(msg, screen,
              (0, VIEWHEIGHT + SPRITEHEIGHT), 0)
    line2str = msg

# ---------------------------------------------------------------------
# showhint() - Fetch and display the next hint for the level.
#
def showhint():
    global currenthint, allhintsshown
    # Does nothing if the level has no hints.
    if len(gamemap.hints) == 0:
        return
    # Hints only cost points the first time they are shown.
    if not allhintsshown:
        player.score += HINTSCORE
    printonline2(gamemap.hints[currenthint])
    currenthint += 1
    # Check if all hints have been displayed, and restart the sequence
    # if so.
    if currenthint == len(gamemap.hints):
        currenthint = 0
        allhintsshown = 1

# ---------------------------------------------------------------------
# dodeath() - Performs checks and cleanups after the player dies.
# Returns:  1 if the player is out of lives, 0 otherwise.
#
def dodeath():
    global enemylist, enemydeathtimes
    player.lives -= 1
    # Update the screen to show the where the player's unfortunate
    # demise took place.
    updateframe()
    if player.lives == 0:
        return 1

    # Reset the player.
    player.posy, player.posx = gamemap.playerstart
    player.checkfall()
    player.creditgold()
    player.digsleft = gamemap.diglimit

    # Have the enemies try to drop their gold where they stand.
    for enemy in enemylist:
        enemy.dropgold()

    # Recount the gold on screen.  This is a safety precaution
    # in case some gold happens to be unaccounted for.
    gamemap.goldleft = 0
    for i in range(MAPHEIGHT):
        for j in range(MAPWIDTH):
            if gamemap.data[i][j] == GOLD and\
               (i, j) not in gamemap.false:
                gamemap.goldleft += 1
    if gamemap.goldleft == 0:
        gamemap.doafter(enemylist)
    # Reset the enemies.  Essentially, they get killed off and
    # replaced by new ones.
    enemydeathtimes = []
    enemylist = []
    for pos in gamemap.enemystart:
        enemylist.append(actor.Enemy(pos[0], pos[1]))
    actor.enemylist = enemylist

    # Restore any one-way gates to their original position.  This is
    # needed because in some circumstances the level would become
    # unfinishable otherwise.
    if gamemap.oddflips:
        gamemap.doswitch()
    # Regenerate all digged floors.
    for floor in gamemap.diggedfloors:
        gamemap.data[floor[0]][floor[1]] = FLOOR
    gamemap.diggedfloors = []

    # Start the level over again.
    levelstartscreen()
    gamemap.draw()
    printonline2()           # Restore previous hint, if any.
    return 0

# =====================================================================

# EXECUTION CONTINUES HERE

player = None                # Our hero!

# Outer game loop.  This simply plays the levels in the adventure.
for lev in levelfiles:
    curlevel += 1

    # Create the map for this level.
    gamemap = map.Map(screen, lev)
    actor.gamemap = gamemap

    # Create the player if we need to.
    if not player:
        player = actor.Player()
        actor.player = player

    # Setup the player.
    player.posy, player.posx = gamemap.playerstart
    player.digsleft = gamemap.diglimit

    # Setup the enemies.
    enemylist = []                     # The list of enemies.
    for pos in gamemap.enemystart:
        enemylist.append(actor.Enemy(pos[0], pos[1]))
    actor.enemylist = enemylist

    player.checkfall()
    player.creditgold()

    # Load the player bitmaps.  We have to do this for each new level
    # because the artwork set may change.

    # The animation dictionary is keyed by action (the ACTION..
    # constants) and consists of sequences sprite surfaces.  The length
    # of each sequence must be equal to ANIMSPERFRAME.
    playeranims = {}                   # Player animations.
    for action in range(NUMACTIONS):
        playeranims[action] = []
        for filename in playerfiles[action]:
            surf = pygame.image.load(os.path.join(gamemap.artwork, filename))
            #Err
            surf.set_colorkey(TRANSTUPLE, RLEACCEL)
            playeranims[action].append(surf.convert())

    # Load the enemy bitmaps.
    enemyanims = {}                    # Enemy animations.  See playeranims.
    for action in range(NUMACTIONS):
        enemyanims[action] = []
        for filename in enemyfiles[action]:
            surf = pygame.image.load(os.path.join(gamemap.artwork, filename))
            #Err
            surf.set_colorkey(TRANSTUPLE, RLEACCEL)
            enemyanims[action].append(surf.convert())

    # Start the level.
    levelstartscreen()
    if skiplvl == 1:
        continue

    currenthint = 0             # The currently displayed hint.
    allhintsshown = 0           # Flag so we know when hints repeat.
    # The game is timed using the frame counter.  Each frame is the
    # game's unit of time.  All events will be timed using the frame
    # counter (rather than actual time) so that they will each be
    # equally affected by system slowdowns/speedups.
    framecount = 0              # The frame counter.
    # Within each frame, there are several animations.  Only when the
    # animation cycle finishes can the player move.  (My use of the
    # word "frame" is kind of confusing - see const.py for an
    # explanation.)
    animcount = 0               # The animation counter.
    enemydeathtimes = []        # The enemies' death certificates.
    paused = 0                  # Flag if the user's pressed pause.

    lastframetime = 0           # Time of last framecount increase.

    # Display the map.
    printonline2(gamemap.meta)
    gamemap.draw()

    # Inner game loop.  This handles input, checks for deaths, etc.
    while 1:
        # To be well-behaved, we respond to any requests to quit the
        # program, which may come from the operating system.
        event = pygame.event.poll()
        if event.type == QUIT:
            sys.exit(0)
        if event.type == ACTIVEEVENT:
            ## Should pause if we're no longer active, and redraw the
            ## entire screen if we are.
            pass

        # PyRunner uses the keyboard state to handle multiple keys at
        # once (e.g. run and dig).
        ## Keystrokes can be missed in some cases.  Fix this somehow.
        keystate = pygame.key.get_pressed()
        ## To handle joystick input, probably the easiest thing to do
        ## is to convert it to keystrokes (e.g. if moving left, then set
        ## keystate[K_LEFT] = 1), although the smarter thing to do would
        ## be to create events (e.g. MOVELEFT) based on either.
        if keystate[K_F5]:         # Player gave up.
            playerdead = 1
            break
        if keystate[K_F6]:              # Player wants a hint.
            showhint()
            updateframe()
            pygame.time.delay(KEYLATENCYDELAY)
        if keystate[K_ESCAPE]:              # Player needs a break.
            paused = not paused
            if paused:
                font.blit("* Game Paused *".center(80), screen, (0, 14))
            else:
                # Redrawing the whole map is overkill, but it doesn't
                # hurt anything here.
                gamemap.draw()
            updateframe()
            pygame.time.delay(KEYLATENCYDELAY)

        # Check if it's time to move to the next frame.
        if not paused and \
           pygame.time.get_ticks() >= lastframetime + FRAMEDELAY:
            # Move to the next animation.
            animcount += 1
            lastframetime = pygame.time.get_ticks()
            for obj in [x for x in enemylist + [player] if x.moving]:
                if dirdict[obj.lastaction] == DIRLEFT:
                    obj.ofsx -= ANIMSTEPX
                if dirdict[obj.lastaction] == DIRRIGHT:
                    obj.ofsx += ANIMSTEPX
                if dirdict[obj.lastaction] == DIRUP:
                    obj.ofsy -= ANIMSTEPY
                if dirdict[obj.lastaction] == DIRDOWN:
                    obj.ofsy += ANIMSTEPY
            if animcount == ANIMSPERFRAME:
                animcount = 0
                player.moving = 0
                for enemy in enemylist:
                    enemy.moving = 0

                framecount += 1

                # Erase the player.
                gamemap.draw(player.posy, player.posx)

                player.checkfall()
                player.creditgold()

                if not player.falling:
                    # Handle input.
                    # NOTE:  To customize the controls, change the K_xxx
                    # constants below.
                    ## Key constants should be aliased (e.g. KEY_MOVELEFT)
                    ## so they can be reconfigured inside the program.
                    if keystate[K_LEFT]: player.moveleft()
                    if keystate[K_RIGHT]: player.moveright()
                    if keystate[K_UP]: player.moveup()
                    if keystate[K_DOWN]: player.movedown()
                    if keystate[K_F8]: player.digleft(framecount)
                    if keystate[K_F7]: player.digright(framecount)
                else:
                    # Player can't move/dig while falling.
                    player.movedown()
                player.creditgold()

                # Check if the player died.  (The R key lets the player
                # kill herself/himself if stuck.)
                if player.checkdeath() or keystate[K_F4]:
                    if dodeath():
                        playerdead = 1
                        break

                # Move the enemies.
                ## BUG:  It's possible for an enemy to walk "on top of but
                ## on square left/right of" another enemy (i.e. in mid air).
                ## This is a logic bug that may have to do with when
                ## checkfall() is called.
                for enemy in enemylist:
                    enemy.checkfall()
                    enemy.domove()
                    if enemy.checkdeath():
                        enemydeathtimes.append(framecount)
                        if gamemap.data[enemy.posy][enemy.posx] == FLOOR:
                            player.score += KILLSCORE
                        else:
                            player.score += LUCKYKILLSCORE
                        enemylist.remove(enemy)
                # Recheck the falling positions.  This is needed so that,
                # for example, if enemy 4 moves out from underneath enemy
                # 0, enemy 0 will then be able to fall.
                for enemy in enemylist:
                    enemy.checkfall()
                # Recheck the player's fall condition for the same reason.
                player.checkfall()

                # We check the player's death again in case an enemy moved
                # onto the player's position.
                if player.checkdeath():
                    if dodeath():
                        playerdead = 1
                        break

                # See if the level is completed.
                if player.checkexit():
                    player.score += player.lives * LIFEBONUSSCORE
                    if curlevel == len(levelfiles):
                        player.score += WINGAMESCORE
                    break

                for obj in [x for x in enemylist + [player] if x.moving]:
                    if dirdict[obj.lastaction] == DIRLEFT:
                        obj.ofsx = SPRITEWIDTH
                    if dirdict[obj.lastaction] == DIRRIGHT:
                        obj.ofsx = -SPRITEWIDTH
                    if dirdict[obj.lastaction] == DIRUP:
                        obj.ofsy = SPRITEHEIGHT
                    if dirdict[obj.lastaction] == DIRDOWN:
                        obj.ofsy = -SPRITEHEIGHT

                checktimedevents()
                updateframe()
            else:
                # No move, only an animation frame.
                updateframe()
        # End of inner game loop.

    # This is a courtesy screen update so the map shows the player
    # actually in the exit point.
    updateframe()

    # Stop the loop if the player has no more lives or pressed Esc to
    # quit.
    if playerdead:
        break
# End of outer game loop.

# Game Over / You Win display.
## Should be positioned using named constants.
screen.fill((0, 0, 255), ((SCREENWIDTH / 2) - 160,
                          (SCREENHEIGHT / 2) - 60,
                          320, 120))
if playerdead:
    msg = "*** GAME OVER ***"
else:
    # Player must have finished all the levels.
    msg = "!! You Win !!"
font.blit(msg.center(40), screen,
          ((SCREENWIDTH / 2) - 160 + (font.width / 2),
           (SCREENHEIGHT / 2) - (font.height * 2)), 0)
font.blit("Press any key to continue".center(40), screen,
          ((SCREENWIDTH / 2) - 160 + (font.width / 2),
           (SCREENHEIGHT / 2) + font.height), 0)
pygame.display.flip()
# Purge any incoming keys so the player can't accidentally skip past
# this display.
pygame.event.get([KEYDOWN])
# Wait for a keypress.
pygame.time.delay(600)
#while pygame.event.poll().type != KEYDOWN: pass

# High scores & post-game cleanup.

# High scores are stored as a list of objects of type highs.HighScore
# and saved/loaded to disk using (non-binary) cPickle.  The list is
# sorted in descending order, so the best score is in index 0.
# High scores is kind of a misnomer, because EVERY score is saved.  (I'm
# a statistics freak, so what did you expect?)

# Construct the high score file name.
highscorefile = gamefilename.split('.')[0] + ".sco"
highscores = []              # The list of high score entries.

try:
    f = open(highscorefile, "r")
    highscores = cPickle.load(f)
except IOError:
    ## Should test errno from the exception to make sure it was a
    ## file not found error.
    # Probably means the file doesn't exist, which is fine - it means
    # the high scores will be blank.
    pass
f.close()

hs = highs.HighScore()       # The high score entry for this game.
hs.score = player.score
hs.level = curlevel
hs.name = playername
hs.time = time.time()

# Thanks to Python lists, maintaining the sorted high scores takes all
# of two lines.
highscores.append(hs)
highscores.sort()

# Save the updated high scores.
f = open(highscorefile, "w")
cPickle.dump(highscores, f)
f.close()

# Find the player's ranking for this game.
ouridx = highscores.index(hs)
#Err - Actually an assert, since we should always find hs.

# The high score screen.
## All of the positioning here should be done using named constants.
screen.fill((0, 0, 0))

# Always show the 5 best scores.
font.blit("PyRunner Top 5".center(80), screen, (0, 0))
for i in range(5):
    if i >= len(highscores):
        break
    font.blit("   %4i.  " % (i + 1) + str(highscores[i]), screen, (0, i + 2))
    # If the player made the top 5, highlight their entry.
    if i == ouridx:
        font.blit(">>", screen, (0, i + 2))
        font.blit("<<", screen, (78, i + 2))

# Show the rankings close to where the player finished.  The basic
# premise here is that the player's entry should appear in the middle
# of the 15-score display, but if the player is at the top or bottom
# then we adjust to show as many scores as possible.
startidx = ouridx - 7
if ouridx + 7 >= len(highscores): startidx = len(highscores) - 15
if startidx < 0: startidx = 0

font.blit("Your Ranking".center(80), screen, (0, 8))
font.blit("   Rank   Name                              Score   "
          "Level  Date & Time", screen, (0, 10))

# Show the scores near the player's position.
for i in range(15):
    if startidx + i >= len(highscores):
        break
    font.blit("   %4i.  " % (startidx + i + 1) + str(highscores[startidx + i]),
                          screen, (0, i + 12))
    # Highlight the player's entry.
    if startidx + i == ouridx:
        font.blit(">>", screen, (0, i + 12))
        font.blit("<<", screen, (78, i + 12))

# Compute some statistics.
avgsc = 0.0                  # Average score.
avglv = 0.0                  # Average level.
for h in highscores:
    avgsc += h.score
    avglv += h.level
avgsc /= len(highscores)
avglv /= len(highscores)

# Show the stats.
tmpstr = "Games Played: %i   Average Score: %.3f   Average Level: %.3f" %\
         (len(highscores), avgsc, avglv)
font.blit(tmpstr.center(80), screen, (0, 28))

font.blit("-- Press any key to exit --".center(80), screen, (0, 29))

pygame.display.flip()

# Purge the keyboard buffer, then wait for a key.
pygame.event.get([KEYDOWN])
while pygame.event.poll().type != KEYDOWN: pass

## Should loop back and let the player play again if desired.

# All done!  Python and pygame will cleanup for us on exit.

# End of file:  pyrunner.py
